"""
CBMPy: CBCommon module
======================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBCommon.py 660 2018-09-24 14:57:04Z bgoli $)

"""
## gets rid of "invalid variable name" info
# pylint: disable=C0103
## gets rid of "line to long" info
# pylint: disable=C0301
## use with caution: gets rid of module xxx has no member errors (run once enabled)
# pylint: disable=E1101


# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import weakref, ast

# this is a hack that needs to be streamlined a bit
#try:
    #import cStringIO as csio
#except ImportError:
    #import io as csio

chemElementDefs = None
HAVE_PYPARSING = False
try:
    from . import pyparsing
    HAVE_PYPARSING = True
except ImportError:
    try:
        print('Attempting global PyParsing import')
        import pyparsing
        HAVE_PYPARSING = True
    except ImportError:
        print('INFO please install pyparsing to enable chemical balance checking!')

if HAVE_PYPARSING:
    pp_caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    pp_lowers = pp_caps.lower()
    pp_digits = "0123456789"
    pp_element = pyparsing.Word(pp_caps, pp_lowers)
    pp_integer = pyparsing.Word(pp_digits)
    pp_elementRef = pyparsing.Group(pp_element + pyparsing.Optional(pp_integer, default="1"))
    pp_chemicalFormula = pyparsing.OneOrMore(pp_elementRef)


    pp_gene_ass = pyparsing.Word(pyparsing.alphanums + '. ') | 'or' | 'OR' | 'and' | 'AND'
    pp_gene_unnester = pyparsing.nestedExpr(content=pp_gene_ass)

ptElements = ('H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',\
              'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'G',\
              'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag',\
              'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Hf', 'Ta', 'W', 'Re', 'Os',\
              'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Rf', 'Db',\
              'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', '112', '113', '114', '115', '116', '117', '118', 'La',\
              'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Ac',\
              'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'X', 'R', 'Z')

## too crude, need to find a better logging strategy
#CBMPYLOG = csio.StringIO()
#def printl(st, scr=True, log=False):
    #"""
    #Updated print() function that also writes to a csio buffer

    #- *st* the string
    #- *prn* print to screen
    #- *log* write to log buffer

    #"""
    #global CBMPYLOG
    #if scr:
        #print(st)
    #if log:
        #try:
            #CBMPYLOG.write(st+'\n')
        #except Exception as ex:
            #print(ex)

#def flushLogToopen(fname):
    #global CBMPYLOG
    #F = open(fname, 'w')
    #CBMPYLOG.seek(0)
    #F.write(CBMPYLOG.read())
    #F.flush()
    #F.close()
    #CBMPYLOG.close()
    #CBMPYLOG = csio.StringIO()


def parseGeneAssociation(gs):
    """
    Parse a COBRA style gene association into a nested list.

     - *gs* a string containing a gene association

    """
    # gs = '(b0810) and ( b0811 ) or ( b1234.0) and(b0809)and ( b7643 )OR(b0812 )AND( b0876)'
    gl = pp_gene_unnester.parseString('('+gs+')').asList()
    gl = gl[0]

    def f(l):
        for i in range(len(l)):
            try:
                l[i] = l[i].lower().strip()
            except Exception:
                f(l[i])

    f(gl)
    print(gl)
    return gl

def checkChemFormula(cf, quiet=False):
    """
    Checks whether a string conforms to a Chemical Formula C3Br5 etc, returns True/False. Please see the SBML
    Level 3 specification and http://wikipedia.org/wiki/Hill_system for more information.

     - *cf* a string that contains a formula to check
     - *quiet* [default=False] do not print error messages

    """
    if not HAVE_PYPARSING:
        print('\nChemical Formula parser requires PyParsing!')
        return False
    try:
        R2 = [(r[0], int(r[1])) for r in pp_chemicalFormula.parseString(cf)]
    except Exception:
        if not quiet:
            print('WARNING: \"{}\" is not a valid chemical formula.'.format(cf))
        return False

    cf2 = ''
    for l_ in R2:
        if l_[0] not in ptElements:
            #print(R2)
            if not quiet:
                print('WARNING: \"{}\" is not a valid chemical formula.'.format(cf))
            return False
        else:
            cf2 += l_[0]
            if l_[1] != 1:
                cf2 += str(l_[1])

    if cf != cf2:
        if not quiet:
            print(R2)
            print('INFO: \"{}\" check formula.'.format(cf))
        return False
    else:
        return True

def extractGeneIdsFromString(g, return_clean_gpr=False):
    """
    Extract and return a list of gene names from a gene association string formulation

    - *g* a COBRA style gene association string
    - *return_clean_gpr* [default=False] in addition to the list returns the "cleaned" GPR string

    """
    #print('\n{}'.format(g))
    g2 = g
    if 'AND' in g2 or 'and' in g2:
        g2 = g2.replace(')AND ', ') AND ').replace(')and ', ') and ')
        g2 = g2.replace(' AND(', ' AND (').replace(' and(', ' and (')
        g2 = g2.replace(')AND(', ') AND (').replace(')and(', ') and (')
    if 'OR' in g2 or 'or' in g2:
        g2 = g2.replace(')OR ', ') OR ').replace(')or ', ') or ')
        g2 = g2.replace(' OR(', ' OR (').replace(' or(', ' or (')
        g2 = g2.replace(')OR(', ') OR (').replace(')or(', ') or (')

    g2 = g2.replace(' AND ', ' and ').replace(' OR ', ' or ')
    g3 = g2.replace('(', '').replace(')', '')
    tmp = []

    if ' and ' in g3 or ' or ' in g3:
        if ' and ' and not ' or ' in g3:
            tmp = g3.split(' and ')
        elif ' or ' and not ' and ' in g3:
            tmp = g3.split(' or ')
        else:
            tmp1 = g3.split(' or ')
            tmp = []
            for s_ in tmp1:
                if ' and ' in s_:
                    tmp3 = s_.split(' and ')
                    tmp += tmp3
                else:
                    tmp += [s_]
    else:
        tmp = [g3]
    names = []
    for n_ in tmp:
        n = n_.strip()
        if n not in names:
            names.append(n)
    if names == ['']:
        names = []

    if not return_clean_gpr:
        return names
    else:
        return names, g2

def createAssociationDictFromNode(node, out, model, useweakref=True, cntr=0):
    """
    Converts a GPR string '((g1 and g2) or g3)' to a dictionary via a Python AST.
    In future I will get rid of all the string elements and work only with AST's.

     - *node* a Python AST node (e.g. body)
     - *out* a gpr dictionary
     - *model* a CBMPy model instance
     - *useweakref* [default=True] use a weakref as the gene object or alternatively the label

    """
    if isinstance(node, ast.Name):
        gene = model.getGene(node.id)
        if gene is not None:
            if useweakref:
                out[node.id] = weakref.ref(gene)
            else:
                out[node.id] = gene.getLabel()
        del gene

    elif isinstance(node, ast.BinOp):
        left = node.left.id
        right = node.right.id
        gref = '{}-{}'.format(left, right)
        print(gref)
        #ref.setGeneProduct(formatSbmlId(gref))
    else:
        if isinstance(node, ast.Expr):
            children = [node.value]
        else:
            children = node.values
        for v in children:
            if isinstance(v, ast.BoolOp) and isinstance(v.op, ast.And):
                key = '_AND_{}'.format(cntr)
                out[key] = {}
                newex = out[key]
                cntr += 1
                #print('And', v)
                #walk(v, newand)
            elif isinstance(v, ast.BoolOp) and isinstance(v.op, ast.Or):
                key = '_OR_{}'.format(cntr)
                out[key] = {}
                newex = out[key]
                cntr += 1
                #print('Or', v)
                #walk(v, newor)
            else:
                #print('-->', v)
                newex = out
            createAssociationDictFromNode(v, newex, model, useweakref=useweakref, cntr=cntr)


def func_getAssociationStrFromGprDict(gprd, out, parent=''):
    """
    Get a old school GPR association string from a CBMPy gprDict, e.g. obtained from gpr.getTree()

     - *gprd* the gprDictionary
     - *out* the output string
     - *parent* [default=''] the string representing the current nodes parent relationship, used for recursion

    """
    out2 = '('
    for k in gprd:
        if k.startswith('_AND_'):
            out2 += '{}{}'.format(func_getAssociationStrFromGprDict(gprd[k], out, ' and '), parent)
        elif k.startswith('_OR_'):
            out2 += '{}{}'.format(func_getAssociationStrFromGprDict(gprd[k], out, ' or '), parent)
        else:
            out2 += '{}{}'.format(k, parent)
    if out2.endswith(' and '):
        out2 = out2[:-5]
    elif out2.endswith(' or '):
        out2 = out2[:-4]
    out2 += ')'
    out = out + out2
    if ' and ' not in out and ' or ' not in out:
        out = out[1:-1]
    return out

# '( llmg_1896  and  llmg_1897  and  llmg_1898  and  llmg_1899  and  llmg_1900  and  llmg_1901 )'
def getGPRasDictFromString(node, out):
    """
    Converts a GPR string '((g1 and g2) or g3)' to a gprDict which is returned

     - *node* a Python AST note (e.g. `ast.parse(gprstring).body[0]`)
     - *out* a new dictionary that will be be created in place

    """

    if isinstance(node, ast.Name):
        #print('Gene: {}'.format(node.id))
        out[node.id] = node.id
    elif isinstance(node, str):
        out[node] = node
    elif isinstance(node, ast.Str):
        out[node.s] = node.s
    elif isinstance(node, ast.BinOp):
        left = node.left.id
        right = node.right.id
        gref = '{}-{}'.format(left, right)
        #print('BinOp: {}'.format(gref))
    else:
        if isinstance(node, ast.Expr):
            children = [node.value]
        elif isinstance(node, ast.BoolOp):
            children = node.values
        else:
            print('Childless node: ', type(node))
            children = []
        cntr2 = 0
        for v in children:
            if isinstance(v, ast.BoolOp) and isinstance(v.op, ast.And):
                out['_AND_{}'.format(cntr2)] = {}
                #print('And', v)
                getGPRasDictFromString(v, out['_AND_{}'.format(cntr2)])
            elif isinstance(v, ast.BoolOp) and isinstance(v.op, ast.Or):
                out['_OR_{}'.format(cntr2)] = {}
                #print('Or', v)
                getGPRasDictFromString(v, out['_OR_{}'.format(cntr2)])
            else:
                #print('-->', v)
                getGPRasDictFromString(v, out)
            cntr2 += 1
    return out


class ComboGen(object):
    """
    Generate sets of unique combinations

    """
    combo = None
    combo_int = None

    def __init__(self):
        self.combo = []

    def addCombination(self, data):
        self.combo.append(','.join(data))

    def uniqueCombinations(self, data, number, temp=None):
        if temp == None:
            temp = []
        if not number:
            self.addCombination(temp)
        for i in range(len(data)):
            temp.append(data[i])
            self.uniqueCombinations(data[i+1:], number-1, temp)
            temp.pop()

    def numberifyComb2Int(self):
        tmp = []
        for c in self.combo:
            tmp.append(tuple([int(c) for c in c.split(',')]))
        self.combo_int = tmp

def processSpeciesChargeChemFormulaAnnot(s, getFromName=False, overwriteChemFormula=False, overwriteCharge=False):
    """
    Disambiguate the chemical formula from either the Notes or the overloaded name

     - *s* a species object
     - *getFromName* [default=False] whether to try strip the chemical formula from the name (old COBRA style)
     - *overwriteChemFormula* [default=False]
     - *overwriteCharge* [default=False]

    """
    tempF = ''
    if overwriteChemFormula:
        tempF = s.chemFormula
        s.chemFormula = None
    key = None
    if s.chemFormula == None or s.chemFormula == '':
        if 'chemFormula' in s.annotation:
            key = 'chemFormula'
        elif 'FORMULA' in s.annotation:
            key = 'FORMULA'
        if key != None:
            s.chemFormula = s.annotation[key]

    if not checkChemFormula(s.chemFormula, quiet=True):
        if getFromName:
            try:
                n = s.getName()
                n, cf = n.rsplit('_', 1)
                if checkChemFormula(cf, quiet=True):
                    s.chemFormula = cf
                    s.setName(n)
                else:
                    s.chemFormula = ''
            except Exception:
                s.chemFormula = ''
        else:
            s.chemFormula = ''
        if overwriteChemFormula and s.chemFormula == '':
            s.chemFormula = tempF

    else:
        if key != None:
            s.annotation.pop(key)

    key2 = None
    tempC = None
    if overwriteCharge:
        tempC = s.charge
        s.charge = None
    if s.charge == None or s.charge == '':
        if 'charge' in s.annotation:
            key2 = 'charge'
        elif 'CHARGE' in s.annotation:
            key2 = 'CHARGE'
        if key2 != None:
            chrg = s.annotation[key2]
            try:
                s.charge = int(chrg)
            except ValueError:
                print('Invalid charge: {} defined for species {}'.format(chrg, s.getId()))
                s.charge = None
                key2 = None
    if overwriteCharge and s.charge == None:
        s.charge = tempC
    if key2 != None:
        s.annotation.pop(key2)


def binHash(keys, d):
    out = []
    for k in keys:
        if k in d:
            out.append(d[k])
        else:
            out.append(True)
    return tuple(out)

def fixId(s, replace=None):
    """
    Checks a string (Sid) to see if it is a valid C style variable. first letter must be an underscore or letter,
    the rest should be alphanumeric or underscore.

    - *s* the string to test
    - *replace* [None] default is to leave out offensive character, otherwise replace with this one

    """

    if s[0].isdigit():
        s2 = '_'
        #print('Illegal start character \"{}\": prefixing an underscore'.format(s[0]))
    else:
        s2 = ''
    for c in s:
        if c.isalnum() or c == '_':
            s2 += c
        elif replace != None:
            s2 += replace
            #print('Replacing illegal character \"{}\": {} --> {}'.format(c, s, s2))
        #else:
            #print('Skipping illegal character \"{}\": {} --> {}'.format(c, s, s2))
    return s2

def checkId(s):
    """
    Checks the validity of the string to see if it conforms to a C variable. Returns true/false

    - *s* a string

    """
    cntr = 0
    for c in s:
        if cntr == 0 and c.isalpha() or c == '_':
            pass
        elif cntr > 0 and c.isalnum() or c == '_':
            pass
        else:
            print('\"{}\" is an invalid character in \"{}\"'.format(c, s))
            return False
        cntr += 1
    return True



# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import os, nose
import cbmpy

fDir = os.path.split(__file__)[0]
#fDir = os.path.split(fDir)[0]
print('TestBase: ', fDir)

def run():
    cDir = os.getcwd()
    os.chdir(fDir)
    res = nose.run()
    os.chdir(cDir)
    if res:
        print('\nAll tests passed.\n')
    else:
        print('\nPlease ignore the second test failure:\n\n\"ERROR: Failure: ImportError (No module named cbmpy)\"\n\nthis is a known glitch in the testing framework that is currently being worked on.\n')

class TestBase:
    def test_import(self):
        import cbmpy



"""
CBMPy: __init__.py
==================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: __init__.py 660 2018-09-24 14:57:04Z bgoli $)

"""
#
# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import os

# Setup some environment
__SILENT_START__ = False
from . import CBConfig
__CBCONFIG__ = CBConfig.__CBCONFIG__
__version__ = CBConfig.__CBCONFIG__['VERSION']
__CBCONFIG__['CBMPY_DIR'] = os.path.split(CBConfig.__file__)[0]

# This is just a hack for backwards compatibility with existing scripts
# that enables "from cbmpy import *" to work - bgoli

# lets add some shortcuts that will be core commands
__HAVE_SOLVER__ = False
try:
    from . import CBSolver
    analyzeModel = CBSolver.analyzeModel
    FluxVariabilityAnalysis = CBSolver.FluxVariabilityAnalysis
    MinimizeSumOfAbsFluxes = CBSolver.MinimizeSumOfAbsFluxes
    doFBA = CBSolver.analyzeModel
    doFVA = CBSolver.FluxVariabilityAnalysis
    doFBAMinSum = CBSolver.MinimizeSumOfAbsFluxes
    __HAVE_SOLVER__ = True
except (ImportError, AttributeError):
    print('No solver present, unable to create shortcuts')

from .CBWrite import writeFVAtoCSV, writeModelToExcel97
from .CBXML import _HAVE_SBML_
if _HAVE_SBML_:
    from .CBXML import sbml_fileFindVersion, sbml_fileValidate
    from .CBRead import loadModel, readSBML3FBC, readSBML2FBA, readCOBRASBML
    from .CBWrite import saveModel, writeSBML3FBC, writeModelToExcel97, writeModelToCOMBINEarchive, writeCOBRASBML, writeSBML3FBCV2
else:
    print('\nWARNING: No SBML support, top-level SBML read/write functions disabled.\n')

# CBMPy GUI development has ended with WxGUI, download Metatoolkit instead
## try:
    ## from .CBWx import runModelEditor as loadCBGUI
## except ImportError as ex:
    ## print('GUI module not available')

##import CBMultiEnv to be deprecated
#try:
    #from . import CBMultiEnv
#except ImportError as ex:
    #print(ex)
    #print('MultiEnvironment module not available')
    #CBMultiEnv = None
#import CBMultiCore
try:
    from . import CBMultiCore
except ImportError as ex:
    print(ex)
    print('MultiCore module not available')
    CBMultiCore = None

INF = float('inf')
NINF = -float('inf')
NAN = float('nan')

from . import CBModel, CBDataStruct, CBModelTools, CBRead, CBReadtxt, CBTools, CBWrite, CBXML, CBNetDB, CBPlot, CBMultiModel
from .CBTools import deSerialize, deSerializeFromDisk
from . import PyscesSED
SED = PyscesSED.SEDCBMPY

try:
    __test_dir__ = os.path.join(__CBCONFIG__['CBMPY_DIR'], 'nosetests')
    test_set = 1
    test_bundle = os.path.join(__test_dir__, 'cbmpy_test_bundle_v{}.zip.py'.format(test_set))
    test_file = os.path.join(__test_dir__, 'installed.v{}'.format(test_set))
    if not os.path.exists(test_file):
        import zipfile
        print('Installing test files (v{})...'.format(test_set))
        zfile = zipfile.ZipFile(test_bundle, allowZip64=True)
        zfile.extractall(path=__test_dir__)
        zfile.close()
        #os.remove(test_bundle)
        del zipfile, zfile
    from . import nosetests
    test = nosetests.run
    del nosetests, test_bundle, test_file, test_set
except ImportError:
    def test():
        print("ERROR: Tests not installed")

del os

if not __SILENT_START__:
    print('\nCBMPy environment\n******************')
    rev = CBConfig.__CBCONFIG__['VERSION_STATUS']
    #print('Release: {}'.format(__version__.replace('.{}'.format(rev),'')))
    print('Revision: r{}'.format(CBConfig.__CBCONFIG__['VERSION_STATUS']))
    #print 'pysces.model_dir = ' + model_dir
    #print 'pysces.output_dir = ' + output_dir

    print('\n\n***********************************************************************')
    print('* Welcome to CBMPy ({}) - PySCeS Constraint Based Modelling       *'.format(__version__.replace('.{}'.format(rev),'')))
    print('*                http://cbmpy.sourceforge.net                         *')
    print('* Copyright(C) Brett G. Olivier 2014 - 2018                           *')
    print('* Dept. of Systems Bioinformatics                                     *')
    print('* Vrije Universiteit Amsterdam, Amsterdam, The Netherlands            *')
    print('* CBMPy is developed as part of the BeBasic MetaToolKit Project       *')
    #print('* PySCeS is distributed under the PySCeS (BSD style) licence, see     *')
    print('* Distributed under the GNU GPL v 3.0 licence, see                    *')
    print('* LICENCE (supplied with this release) for details                    *')
    ##  print('*                 ** Read about PySCeS **                             *')
    ##  print('* http://bioinformatics.oupjournals.org/cgi/content/short/21/4/560    *')
    print('***********************************************************************\n')




'''
This code is adapted from https://gist.github.com/adewes/5884820

Thanks to adewes for putting it online

TODO: clarify copyright - if I would not be so lazy I could have come up with it
myself
'''

import random
 
myrandom = random.Random(135) 

def get_random_color():
    x = myrandom.uniform(0,1.0)
    y = myrandom.uniform(0.7,1.0)
    z = myrandom.uniform(0.7,1.0)
    return (x, y, z)
 
def color_distance(c1,c2):
    return min(abs(c1[0] - c2[0]), 1-abs(c1[0] - c2[0])) + abs(c1[1] - c2[1]) + abs(c1[2] - c2[2])
 
def generate_new_color(existing_colors):
    """ creates a new random color (in hsv scheme) that is most different
    to the existing colors.
    """
    max_distance = None
    best_color = None
    for i in range(0,100):
        color = get_random_color()
        if not existing_colors:
            return color
        best_distance = min([color_distance(color,c) for c in existing_colors])
        if not max_distance or best_distance > max_distance:
            max_distance = best_distance
            best_color = color
    return best_color
 


"""
CBMPy: __init__.py
==================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2015 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: __init__.py 358 2015-08-11 10:41:01Z bgoli $)

"""

print('PyscesCBM is now developed as CBMPy (cbmpy.sourceforge.net) please update your code:')
print('\n\timport cbmpy\n\timport cbmpy as cbm\n')
print('My apologies for any inconvenience.\n\n')
raise ImportError('Obsolete module please use CBMPy:\t"import cbmpy"')


import os, time, numpy
cDir = os.path.dirname(os.path.abspath(os.sys.argv[0]))
import cbmpy as cbm

print os.sys.argv

mBase = os.sys.argv[2]
testDir = os.path.join(os.sys.argv[1], mBase)
outDir = os.sys.argv[3]

if not os.path.exists(outDir):
    os.makedirs(outDir)

cmod = cbm.CBRead.readSBML3FBC('%s-sbml-l3v1.xml' % mBase, testDir)
## cbm.CBSolver.cplx_analyzeModel(cmod, lpFname=os.path.join(outDir, mBase))
## cbm.CBSolver.glpk_analyzeModel(cmod)
cbm.CBSolver.analyzeModel(cmod)

F = file(os.path.join(testDir, '%s-settings.txt' % mBase), 'r')
variables = None
for l_ in F:
    if l_.startswith('variables:'):
        variables = l_.strip().split(':')[1].strip().split(',')
        break
#print variables
data = []
for v_ in variables:
    vtest = cmod.getReaction(v_)
    if vtest == None:
        data.append(cmod.getActiveObjective().getValue())
    else:
        data.append(cmod.getReaction(v_).getValue())
#print data

F = file(os.path.join(outDir, mBase)+'.csv', 'w')
out = ''
for v_ in variables:
    out += '%s,' % v_
out = '%s\n' % out[:-1]
for d_ in data:
    out += '%s,' % d_
out = '%s\n' % out[:-1]
F.write(out)
F.close()




import os, time, numpy, subprocess, zipfile
cDir = os.path.dirname(os.path.abspath(os.sys.argv[0]))
import cbmpy as cbm


testDir = os.path.join(cDir, 'tests')
if not os.path.exists(testDir):
    os.makedirs(testDir)
    if os.path.exists(os.path.join(cDir, 'fbc_test_suite.zip')):
        print('Unzipping Test suite to: '.format(testDir))
        zip_ref = zipfile.ZipFile(os.path.join(cDir, 'fbc_test_suite.zip'), 'r')
        zip_ref.extractall(testDir)
        zip_ref.close()


outDir = os.path.join(cDir, 'fbc_test_suite_out')
if not os.path.exists(outDir):
    os.makedirs(outDir)

tests = os.listdir(testDir)
print tests

for T in tests:
    retcode = subprocess.call([os.sys.executable, 'run_single_fbc_test.py', testDir, T, outDir])

testResults = []
for T in tests:
    rF = file(os.path.join(testDir, T, '%s-results.csv' % T), 'r')
    r0 = rF.read().strip()
    r0 = [a.split(',') for a in r0.split('\n')]
    r00 = {}
    for j_ in range(len(r0[0])):
        r00[r0[0][j_]] = r0[1][j_]
    rF.close()
    rF2 = file(os.path.join(outDir, '%s.csv' % T), 'r')
    r1 = rF2.read().strip()
    r1 = [a.split(',') for a in r1.split('\n')]
    r11 = {}
    for j_ in range(len(r1[0])):
        r11[r1[0][j_]] = r1[1][j_]
    rF2.close()
    testRes = 'fail'
    for k_ in r00:
        if k_ in r11:
            if r00[k_].upper() == r11[k_].upper():
                testRes = 'pass'
    testResults.append((T, testRes))

print '\n\nFBC Test Suite Results:\n===================='
for t_ in testResults:
    print 'Test %s: %s' % (t_[0], t_[1])



"""
CBMPy: fluxmodules sparse rationals test module
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Arne C. Reimers
Contact email: arne.c.reimers@gmail.com
"""

import unittest

from sparserationals import Matrix

from fractions import Fraction

class Test(unittest.TestCase):


    def setUp(self):
        self.mat = Matrix([[-2, -1,  1,  0,  0,  0,  0,  0,  0],
                           [ 1,  0,  0, -1,  0,  0,  0,  0,  0],
                           [ 0,  1, -1,  1, -1, -1,  0,  0,  0],
                           [ 1,  0,  0,  0,  1,  0, -1,  0,  0],
                           [ 0,  0,  0,  0,  0,  1,  1, -1,  1],
                           [ 0,  0,  0,  0,  0,  0,  0,  1, -1]])


    def tearDown(self):
        pass


    def testPrintNumpy(self):
        self.assertEquals(self.mat.cols(), 9)
        self.assertEquals(self.mat.rows(), 6)
        print(self.mat.toNumpy())

    def testValueAccess(self):
        self.assertEqual(self.mat[0,0], -2)
        self.assertEqual(self.mat[5,3], 0)
        self.assertEquals(self.mat[3,6], -1)
        
    def testSubMatrix(self):
        submat = self.mat[:,[2,4]]
        self.assertEqual(submat[0,0], 1)
        self.assertEqual(submat[1,0], 0)
        self.assertEqual(submat[2,0], -1)
        self.assertEqual(submat[3,0], 0)
        self.assertEqual(submat[4,0], 0)
        self.assertEqual(submat[5,0], 0)
        self.assertEqual(submat[0,1], 0)
        self.assertEqual(submat[1,1], 0)
        self.assertEqual(submat[2,1], -1)
        self.assertEqual(submat[3,1], 1)
        self.assertEqual(submat[4,1], 0)
        self.assertEqual(submat[5,1], 0)
        
    def testTranspose(self):
        t = self.mat.transpose()
        tt = t.transpose()
        tt -= self.mat
        for i in range(tt.rows()):
            for j in range(tt.cols()):
                self.assertEqual(tt[i,j], 0)
    
    def testSetItem(self):
        submat = self.mat[[1,3],1:5]
        c = self.mat.copy()
        submat *= Fraction(-5)
        c[[1,3],1:5] += submat / 5 
        
        for i in [1,3]:
            for j in range(*(1,5)):
                self.assertEqual(c[i,j], 0)
        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


"""
CBMPy: fluxmodules matroid test module

This module only exists for testing purposes
=====================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Arne C. Reimers
Contact email: arne.c.reimers@gmail.com
"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import unittest

#from sympy import Matrix
from .sparserationals import Matrix

from . import matroid

class TestMatroid(unittest.TestCase):
    """ Test methods in matroid.py """

    def setUp(self):
        """ create test matroid """
        mat = Matrix([[-2, -1,  1,  0,  0,  0,  0,  0,  0],
                      [ 1,  0,  0, -1,  0,  0,  0,  0,  0],
                      [ 0,  1, -1,  1, -1, -1,  0,  0,  0],
                      [ 1,  0,  0,  0,  1,  0, -1,  0,  0],
                      [ 0,  0,  0,  0,  0,  1,  1, -1,  1],
                      [ 0,  0,  0,  0,  0,  0,  0,  1, -1]])
        
        elems = [1,2,3,4,5,6,7,8,9]
        self.matroid = matroid.fromMatrix(mat, elems)

    def tearDown(self):
        """ nothing to do"""
        pass


    def testElems(self):
        self.assertEqual(len(self.matroid.elems), 9)
        self.assertEqual(len(self.matroid.basis)+len(self.matroid.nbasis), 
                          len(self.matroid.elems))
        # every basis must have same size of 5
        self.assertEqual(len(self.matroid.basis), 5)
        
    def testBasis(self):
        print(self.matroid.rep.toNumpy())
        self.assertTrue(self.matroid.isIndependent({1,2,4,6,8}))
        self.assertSetEqual(set(self.matroid.nbasis), {3,5,7,9})
        self.assertSetEqual(set(self.matroid.basis), {1,2,4,6,8})
        
        self.assertTrue(self.matroid.isIndependent({1,3,5,9,7}))
        self.assertTrue(self.matroid.isCoIndependent({3,7,9}))
        
        self.assertFalse(self.matroid.isIndependent({2,3}))
        self.assertTrue(self.matroid.isIndependent({1,4,5}))
        self.assertFalse(self.matroid.isIndependent({8,9}))
        self.assertFalse(self.matroid.isIndependent({1,3,6,7,4}))
    
    def testFundamentalCircuit(self):
        # first set basis
        self.assertTrue(self.matroid.isIndependent({1,2,4,6,8}))
        self.assertSetEqual(self.matroid.fundamentalCircuit(3), {2})
        self.assertSetEqual(self.matroid.fundamentalCircuit(5), {1,2,4})
        self.assertSetEqual(self.matroid.fundamentalCircuit(7), {1,2,4,6})
        self.assertDictEqual(self.matroid.fundamentalCircuitValues(3), {2:1})
        self.assertDictEqual(self.matroid.fundamentalCircuitValues(5), {1:-1, 2:2, 4:-1})
        self.assertDictEqual(self.matroid.fundamentalCircuitValues(7), {1:1, 2:-2, 4:1, 6:-1})

        # if we give basis elements, we get cofundamental circuits
        self.assertSetEqual(self.matroid.fundamentalCircuit(8), {9})
        self.assertSetEqual(self.matroid.fundamentalCircuit(4), {5,7})
 
    def testModules(self):
        modules = self.matroid.findModules()
        self.assertEqual(len(modules), 2)
        for m in modules:
            if len(m.elems) == 2:
                self.assertSetEqual(set(m.elems), {8,9})
            else:
                self.assertSetEqual(set(m.elems), {1,2,3,4,5,6,7})
    
    def testParallel(self):
        parallel = self.matroid.parallel()
        self.assertSetEqual(parallel, {frozenset({2,3}),
                                       frozenset({8,9})})
        
    def testCoParallel(self):
        coparallel = self.matroid.coparallel()
        self.assertSetEqual(coparallel, {frozenset({1,4}),
                                         frozenset({8,9}),
                                         frozenset({6,7})})

    def testDelete(self):
        m = self.matroid.delete({7})
        self.assertFalse(7 in m.elems)
        self.assertTrue(m.isIndependent({6}))
        self.assertFalse(m.isCoIndependent({6}))
        self.assertTrue(m.isIndependent({1,2,4,8}))
        
        self.assertTrue(self.matroid.isCoIndependent({6}))
        
    def testContraction(self):
        m = self.matroid.contract({2,4})
        self.assertFalse(m.isIndependent({3}))
        self.assertFalse(m.isIndependent({1,5}))
        self.assertTrue(m.isIndependent({1}))
        self.assertTrue(m.isIndependent({5}))
        
        parallel = m.contract({6}).parallel()
        self.assertSetEqual(parallel, {frozenset({1,5,7}),
                                       frozenset({8,9})})

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import os, sys, time, zipfile, json
cDir = os.path.dirname(os.path.abspath(__file__))
MDIR = os.path.join(os.path.split(cDir)[0], 'models')
MDIR = cDir

try:
    import nose
except ImportError:
    print('WARNING: To run tests please install python-nose')
from nose.tools import assert_equal
from nose.tools import assert_almost_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import raises

import cbmpy

F = open(os.path.join(MDIR, 'data.json'))
DATA = json.load(F)
F.close()

class TestReadSBML(object):
    @classmethod
    def setup_class(klass):
        """This method is run once for each class before any tests are run"""
        if 'HOME' in os.environ:
            klass.CDIR = os.environ['HOME']
        else:
            klass.CDIR = os.getcwd()

    @classmethod
    def teardown_class(klass):
        """This method is run once for each class _after_ all tests are run"""

    def setUp(self):
        """This method is run once before _each_ test method is executed"""

    def teardown(self):
        """This method is run once after _each_ test method is executed"""

    def test_load_L3FBCV2_iJR904(self):
        cmod = cbmpy.readSBML3FBC(os.path.join(MDIR, 'L3FBCV2_iJR904.glc.xml'))
        assert_not_equal(cmod, None)

    def test_load_L3FBCV1_iJR904(self):
        cmod = cbmpy.readSBML3FBC(os.path.join(MDIR, 'L3FBCV1_iJR904.glc.xml'))
        assert_not_equal(cmod, None)

    def test_load_L2FBA_iJR904(self):
        cmod = cbmpy.readSBML2FBA(os.path.join(MDIR, 'L2FBA_iJR904.glc.xml'))
        assert_not_equal(cmod, None)

    def test_load_L2CBR_iJR904(self):
        cmod = cbmpy.readCOBRASBML('L2CBR_iJR904.glc.xml', work_dir=MDIR,\
                                                output_dir=self.CDIR, delete_intermediate=False)
        assert_not_equal(cmod, None)


class TestRunFBA(object):
    @classmethod
    def setup_class(klass):
        """This method is run once for each class before any tests are run"""
        klass.m = {}
        for m in DATA:
            cmod = None
            if m.startswith('L2CBR_'):
                cmod = cbmpy.readCOBRASBML(m, work_dir=MDIR,\
                                           output_dir=self.CDIR, delete_intermediate=True)
            elif m.startswith('L2FBA_'):
                cmod = cbmpy.readSBML2FBA(os.path.join(MDIR, m))
            elif m.startswith('L3FBCV1_') or m.startswith('L3FBCV2_'):
                cmod = cbmpy.readSBML3FBC(os.path.join(MDIR, m))
            else:
                cmod = None
            if cmod is not None:
                klass.m[m] = cmod
        if 'HOME' in os.environ:
            klass.CDIR = os.environ['HOME']
        else:
            klass.CDIR = os.getcwd()

    @classmethod
    def teardown_class(klass):
        """This method is run once for each class _after_ all tests are run"""
        del klass.m

    def test_load(self):
        for m in self.m:
            assert_not_equal(self.m[m], None)

    def test_run_fba_cobra(self):
        print('\n\nThis test fails under Linux.\n\n')
        cmod = cbmpy.readCOBRASBML('L2CBR_iJR904.glc.xml', work_dir=MDIR,\
                                                        output_dir=self.CDIR, delete_intermediate=True)
        cmod2 = cbmpy.readSBML3FBC('L3FBCV1_iJR904.glc.xml', work_dir=MDIR)
        fba1 = cbmpy.doFBA(cmod)
        fba2 = cbmpy.doFBA(cmod2)
        del cmod, cmod2
        assert_almost_equal(fba1, fba2)

    def test_run_L3FBCV2_iJR904(self):
        m = 'L3FBCV2_iJR904.glc.xml'
        cmod = self.m[m].clone()
        res = cbmpy.doFBA(cmod)
        del cmod
        assert_almost_equal(DATA[m]['FBA'], res)

    def test_run_L3FBCV1_iJR904(self):
        m = 'L3FBCV1_iJR904.glc.xml'
        cmod = self.m[m].clone()
        res = cbmpy.doFBA(cmod)
        del cmod
        assert_almost_equal(DATA[m]['FBA'], res)

    def test_run_L2FBA_iJR904(self):
        m = 'L2FBA_iJR904.glc.xml'
        cmod = self.m[m].clone()
        res = cbmpy.doFBA(cmod)
        del cmod
        assert_almost_equal(DATA[m]['FBA'], res)


class TestBiGG2(object):
    @classmethod
    def setup_class(klass):
        """This method is run once for each class before any tests are run"""
        klass.m = {}
        for m in DATA:
            if m.startswith('BIGG2_'):
                cmod = cbmpy.readSBML3FBC(os.path.join(MDIR, m))
                klass.m[m] = cmod

    @classmethod
    def teardown_class(klass):
        """This method is run once for each class _after_ all tests are run"""
        del klass.m

    def test_load(self):
        for m in self.m:
            assert_not_equal(self.m[m], None)

    def test_run_BIGG2_iIT341(self):
        m = 'BIGG2_iIT341.xml'
        cmod = self.m[m].clone()
        res = cbmpy.doFBA(cmod)
        del cmod
        assert_almost_equal(DATA[m]['FBA'], res)

    def test_run_BIGG2_iAF1260(self):
        m = 'BIGG2_iAF1260.xml'
        cmod = self.m[m].clone()
        res = cbmpy.doFBA(cmod)
        del cmod
        assert_almost_equal(DATA[m]['FBA'], res)




import os, sys, time, zipfile, json
cDir = os.path.dirname(os.path.abspath(__file__))
MDIR = os.path.join(os.path.split(cDir)[0], 'models')
MDIR = cDir

import nose
from nose.tools import assert_equal
from nose.tools import assert_almost_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import raises

import cbmpy

F = open(os.path.join(MDIR, 'data.json'))
DATA = json.load(F)
F.close()

TEMPDATA = {}

class TestReagents(object):
    @classmethod
    def setup_class(klass):
        """This method is run once for each class before any tests are run"""
        klass.mcore = cbmpy.readSBML3FBC('cbmpy_test_core')
        klass.sboterm1 = 'SBO:1234567'
        if 'HOME' in os.environ:
            klass.CDIR = os.environ['HOME']
        else:
            klass.CDIR = os.getcwd()

    @classmethod
    def teardown_class(klass):
        """This method is run once for each class _after_ all tests are run"""
        del klass.mcore

    def roundTripModelV1(self, model):
        modname = 'rtv1.xml'
        cbmpy.writeSBML3FBC(model, modname, directory=self.CDIR)
        del model
        model = cbmpy.readSBML3FBC(modname, work_dir=self.CDIR)
        return model

    def test_param_create(self):
        M = self.mcore.clone()

        pid = 'par1'
        pval = 1.0
        pname = 'parameter1'
        par1 = cbmpy.CBModel.Parameter(pid, pval, pname, True)
        M.addParameter(par1)
        pout = M.getParameter(pid)
        assert pout.getId() == pid
        assert pout.getName() == pname
        assert pout.getValue() == pval
        assert pout.constant == True
        del M

    def test_set_coefficient(self):
        M = self.mcore.clone()

        pval = 10.0
        rid = 'R01'
        suid = 'X0'
        prid = 'A'
        R = M.getReaction(rid)
        sub = R.getReagentWithSpeciesRef(suid)
        prod = R.getReagentWithSpeciesRef(prid)

        sub.setCoefficient(-pval)
        prod.setCoefficient(pval)

        #M = self.roundTripModelV1(M)

        R = M.getReaction(rid)
        sub = R.getReagentWithSpeciesRef(suid)
        prod = R.getReagentWithSpeciesRef(prid)

        assert sub.getCoefficient() == -pval
        assert prod.getCoefficient() == pval

        del M

    def test_set_par_coefficient(self):
        M = self.mcore.clone()

        pname = 'parameter1'
        pval = 10.0
        rid = 'R01'
        suid = 'X0'
        prid = 'A'

        par_sub = cbmpy.CBModel.Parameter('par1', -pval, pname, True)
        par_prod = cbmpy.CBModel.Parameter('par2', pval, pname, True)
        M.addParameter(par_sub)
        M.addParameter(par_prod)

        R = M.getReaction(rid)
        sub = R.getReagentWithSpeciesRef(suid)
        prod = R.getReagentWithSpeciesRef(prid)

        sub.setCoefficient(par_sub)
        prod.setCoefficient(par_prod)

        #M = self.roundTripModelV1(M)

        R = M.getReaction(rid)
        sub = R.getReagentWithSpeciesRef(suid)
        prod = R.getReagentWithSpeciesRef(prid)

        assert sub.getCoefficient() == -pval
        assert prod.getCoefficient() == pval

        del M







    #def test_set_coefficient(self):
        #M = self.mcore.clone()
        #for c in M.reactions:
            #c.setSBOterm(self.sboterm1)
        #M = self.roundTripModelV1(M)
        #for c in M.reactions:
            #assert self.sboterm1 == c.getSBOterm()
        #del M



"""
CBMPy: MultiCoreFVA module
==========================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: _multicorefva.py 660 2018-09-24 14:57:04Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import os, time, math, copy, multiprocessing
try:
    import pickle
except ImportError:
    import cPickle as pickle

import cbmpy

def funcFVA(x,r):
    try:
        Args = x.FVAARGS
        del x.FVAARGS
        Args.insert(0, x)
        Args[1] = r
        fva, fvan = cbmpy.CBSolver.FluxVariabilityAnalysis(*Args)
    except Exception as ex:
        print(ex)
        fva = fvan = None
        raise RuntimeError(ex)
    return fva, fvan

def multiCoreFVA(fvamod, procs=2, timeout=None):
    #assert procs >= 2, '\nMinimum 2 processes required'
    rid = fvamod.getReactionIds()
    PRs = [d_[0] for d_ in list(cbmpy.CBMultiCore.grouper(int(math.ceil(len(rid)/float(procs))), range(len(rid))))]
    print(procs, len(PRs), PRs)

    s2time = time.time()
    TP = multiprocessing.Pool(processes=procs)
    results = []
    if procs == 1:
        results.append(TP.apply_async(funcFVA, (fvamod.clone(), rid)))
    else:
        for p_ in range(len(PRs)-1):
            print('{} -> {}'.format(PRs[p_], PRs[p_+1]))
            print('{} -> {}'.format(rid[PRs[p_]], rid[PRs[p_+1]]))
            # investigate using without clone
            #results.append(TP.apply_async(funcFVA, (fvamod, rid[PRs[p_]:PRs[p_+1]])))
            results.append(TP.apply_async(funcFVA, (fvamod.clone(), rid[PRs[p_]:PRs[p_+1]])))
        results.append(TP.apply_async(funcFVA, (fvamod.clone(), rid[PRs[-1]:])))
    fva = []
    fvan = []
    for r_ in results:
        out = r_.get(timeout)
        fva.append(out[0])
        fvan.append(out[1])
    e2time = time.time()
    del results

    print('\nMulticore FVA took: {} min ({} processes)'.format((e2time-s2time)/60., procs))
    return fva, fvan


if __name__ == '__main__':

    print(os.sys.argv)

    cores = int(os.sys.argv[1])
    F = file(os.sys.argv[2],'rb')
    cmod = pickle.load(F)
    F.close()

    cbmpy.CBSolver.analyzeModel(cmod, oldlpgen=False)
    res = multiCoreFVA(cmod, procs=cores)

    F = file(os.sys.argv[2], 'wb')
    pickle.dump(res, F, protocol=-1)
    F.flush()
    F.close()



import os, sys, time, zipfile, json
cDir = os.path.dirname(os.path.abspath(__file__))
MDIR = os.path.join(os.path.split(cDir)[0], 'models')
MDIR = cDir

import nose
from nose.tools import assert_equal
from nose.tools import assert_almost_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import raises

import cbmpy

F = open(os.path.join(MDIR, 'data.json'))
DATA = json.load(F)
F.close()

TEMPDATA = {}

class TestSBO(object):
    @classmethod
    def setup_class(klass):
        """This method is run once for each class before any tests are run"""
        klass.mcore = cbmpy.readSBML3FBC('cbmpy_test_core')
        klass.sboterm1 = 'SBO:1234567'
        if 'HOME' in os.environ:
            klass.CDIR = os.environ['HOME']
        else:
            klass.CDIR = os.getcwd()

    @classmethod
    def teardown_class(klass):
        """This method is run once for each class _after_ all tests are run"""
        del klass.mcore

    def roundTripModelV1(self, model):
        modname = 'rtv1.xml'
        cbmpy.writeSBML3FBC(model, modname, directory=self.CDIR)
        del model
        model = cbmpy.readSBML3FBC(modname, work_dir=self.CDIR)
        return model

    def test_sbo_reactions_getset(self):
        M = self.mcore.clone()
        for c in M.reactions:
            c.setSBOterm(self.sboterm1)
        for c in M.reactions:
            print(self.sboterm1, c.getSBOterm())
            assert self.sboterm1 == c.getSBOterm()
        del M

    def test_sbo_species_getset(self):
        M = self.mcore.clone()
        for c in M.species:
            c.setSBOterm(self.sboterm1)
        for c in M.species:
            assert self.sboterm1 == c.getSBOterm()
        del M

    def test_sbo_compartments_getset(self):
        M = self.mcore.clone()
        for c in M.compartments:
            c.setSBOterm(self.sboterm1)
        for c in M.compartments:
            assert self.sboterm1 == c.getSBOterm()
        del M

    def test_sbo_reactions_getset_wrt(self):
        M = self.mcore.clone()
        for c in M.reactions:
            c.setSBOterm(self.sboterm1)
        M = self.roundTripModelV1(M)
        for c in M.reactions:
            assert self.sboterm1 == c.getSBOterm()
        del M

    def test_sbo_species_getset_wrt(self):
        M = self.mcore.clone()
        for c in M.species:
            c.setSBOterm(self.sboterm1)
        M = self.roundTripModelV1(M)
        for c in M.species:
            assert self.sboterm1 == c.getSBOterm()
        del M

    def test_sbo_compartments_getset_wrt(self):
        M = self.mcore.clone()
        for c in M.compartments:
            c.setSBOterm(self.sboterm1)
        M = self.roundTripModelV1(M)
        for c in M.compartments:
            assert self.sboterm1 == c.getSBOterm()
        del M



"""
CBMPy: setup.py
===============
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2010-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: setup.py 660 2018-09-24 14:57:04Z bgoli $)

"""

import os
local_path = os.path.dirname(os.path.abspath(os.sys.argv[0]))

try:
    from setuptools import setup
except:
    from distutils.core import setup

mydata_files = []
# new hack should get rid of the need for this
#mydata_files.append((os.path.join('cbmpy', 'models'),\
#[os.path.join(local_path,'src','models',examplefile) for examplefile in os.listdir(os.path.join(local_path,'src','models'))\
#if examplefile.endswith('.xml')]))

# new hack should get rid of the need for this
#mydata_files.append((os.path.join('cbmpy', 'nosetests'),\
#[os.path.join(local_path,'src','nosetests',examplefile) for examplefile in os.listdir(os.path.join(local_path,'src','nosetests'))\
#if examplefile.endswith('.xml') or examplefile.endswith('.json')]))

# release
try:
    STATUS = 'r'+'$Rev: 660 $'.replace('Rev: ','').replace('$','').strip()
except:
    STATUS = 'beta'

mypackages= ['cbmpy', 'cbmpy.fluxmodules', 'cbmpy.models', 'cbmpy.nosetests']#, 'cbmpy.solver']

setup(
 package_dir = {'cbmpy' : 'src'},
 packages = mypackages,
 data_files = mydata_files,
 name = "cbmpy",
 summary = "CBMPy: Constraint Based Modelling in Python",
 #version = "0.7.22.{}".format(STATUS),
 version = "0.7.21",
 maintainer = 'Brett G. Olivier',
 author = 'Brett G. Olivier',
 author_email = 'b.g.olivier@vu.nl',
 maintainer_email = 'bgoli@users.sourceforge.net',
    url = "http://cbmpy.sourceforge.net",
    download_url = "http://sourceforge.net/projects/cbmpy/files/release/0.7/",
    license = "GNU General Public License (GPL)",
    keywords = "computational systems biology, modelling, simulation, genome scale models, sbml, constraint-based modelling, fbc, linear programming, groups, standard" ,
    #zip_safe = False,
    requires = ['numpy', 'sympy', 'libsbml', 'nose'],
    platforms = ["Windows", "Linux", "Mac"],
    classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Chemistry'
     ],
 description =\
 """ CBMPy (http://cbmpy.sourceforge.net) is a platform for constraint based modelling and analysis. CBMPy implements popular analyses such as FBA, FVA, element/charge balancing, network analysis and model editing as well as advanced methods developed specifically for the ecosystem modelling."""
)
try:
    import cbmpy
except ImportError:
    pass

readme =\
 """ PySCeS-CBMPy
 ============

 PySCeS CBMPy (http://cbmpy.sourceforge.net) is a new platform for constraint
 based modelling and analysis. It has been designed using principles developed
 in the PySCeS simulation software project: usability, flexibility and accessibility. Its architecture is both extensible and flexible using data structures that are intuitive to  the biologist (metabolites, reactions, compartments) while transparently translating these into the underlying mathematical structures used in advanced analysis (LP's, MILP's).

 PySCeS CBMPy implements popular analyses such as FBA, FVA, element/charge
 balancing, network analysis and model editing as well as advanced methods
 developed specifically for the ecosystem modelling: minimal distance methods,
 flux minimization and input selection. To cater for a diverse range of modelling
 needs PySCeS CBMPy supports user interaction via:

 - interactive console, scripting for advanced use or as a library for software development
 - GUI, for quick access to a visual representation of the model, analysis methods and annotation tools
 - SOAP based web services: using the Mariner framework much high level functionality is exposed for integration into web tools

 For more information on the development and use of PySCeS CBMPy feel free to contact me:

 PySCeS-CBMPy has been tested on Windows 7 and 8.1, Mac OSX and Ubuntu Linux 12.04, 14.04, 16.04. It is compatible with both Python 2.7+ and includes experimental support for Python 3.4+ It is highly recommend to use
 Python 2.7 as not all Python package dependencies (extended functionality) are available for Python 3.

 PySCeS CBMPy is now accessible as a Python module **cbmpy** in place of the the previously used **pyscescbm** which is no longer supported. CBMPy includes support for  reading/writing models in SBML3 FBC versions 1 and 2 as well as COBRA dialect, Excel spreadsheets and Python.

 To use follow the installation instructions given below and try the following in a Python shell::

 import cbmpy
 cmod = cbmpy.readSBML3FBC('cbmpy_test_core')
 cbmpy.doFBA(cmod)

 New Ipython notebook tutorials are available. Happy modelling!

 The following installation instructions are for Ubuntu 16.04 but should be adaptable to any
 Linux package managment system, OSX, Debian, etc. Except for GLPK (4.47) and SymPy (0.7.4 or newer)
 no specific library version is required. For more detailed installation instructions and Windows
 please see the online documentation http://cbmpy.sourceforge.net/reference/install_doc.html

 New! auto-dependency configuration
 ----------------------------------

 I am in the process of creating automated dependency checking and building tools for CBMPy. These can be found at::

  https://github.com/bgoli/cbmpy-build

 Ubuntu support is almost complete with Windows/Conda support in development, grab form GitHub::

  https://github.com/bgoli/cbmpy-build.git

 Manual dependency configuration is provided below although for almost all users the "batteries included"
 Python distribution Anaconda (http://www.anaconda.io) is recommended. Using conda installing CBMPy
 is then as easy as:

  conda install -c bgoli -c sbmlteam cbmpy

 If you would like to roll your own installation on Linux using the native Python this is also relatively
 easy.

 Python2
 -------

 First we create a scientific Python workbench::

 sudo apt-get install python-dev python-numpy python-scipy python-matplotlib  python-pip
 sudo apt-get install python-sympy python-suds python-xlrd python-xlwt python-h5py
 sudo apt-get install python-wxgtk2.8
 sudo apt-get install ipython ipython-notebook

 libSBML
 ~~~~~~~

 Installing libSBML is now easy using Pip::

 sudo apt-get install libxml2 libxml2-dev
 sudo apt-get install zlib1g zlib1g-dev
 sudo apt-get install bzip2 libbz2-dev

 sudo pip install --update python-libsbml

 Extended functionality
 ~~~~~~~~~~~~~~~~~~~~~~

 sudo pip install biopython docx

 Windows
 ~~~~~~~

 Use easy_install, pip or your package manager (e.g. conda) to install the following packages::

  numpy scipy matplotlib sympy xlrd xlwt
  biopython docx suds wxPython

 pip install --update python-libsbml

 glpk/python-glpk
 ~~~~~~~~~~~~~~~~

 CBMPy requires a linear solver for numerical analysis, the open source (glpk) solver can be automatically built and installed as follows (requires git to be installed and accessible):

 Download the install script that will install GLPK/PyGLPK for CBMPy on Ubuntu 14.04 or newer::

  curl --remote-name https://raw.githubusercontent.com/bgoli/cbmpy-glpk/master/install_glpk.sh

 Make executable::

  chmod 744 install_glpk.sh

 and run::

  ./install_glpk.sh

 Note this script is designed to be used for building containers and will remove any installed version of GLPK and build and install the correct version needed for PyGLPK.

 No warranty of any kind assumed or otherwise, use at own risk!

 CBMPy
 ~~~~~

 Finally, install CBMPy::

  sudo easy_install cbmpy

 or

  sudo pip install cbmpy

 or try the new CONDA support::

  conda install -c bgoli -c sbmlteam cbmpy

 or download the source  and run::

  python setup.py build sdist
  sudo python setup.py install

"""



"""
CBMPy: CBDataStruct module
==========================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBDataStruct.py 660 2018-09-24 14:57:04Z bgoli $)

"""

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

import os, time, copy, re
import numpy

import webbrowser
from . import miriamids

MIRIAM = miriamids.miriamids

#MIRIAM.update({'EC' : {'data_entry': 'http://www.ebi.ac.uk/intenz/query?cmd=SearchEC&amp;ec=$id',
                       #'example': '1.1.1.1',
                       #'name': 'Enzyme Nomenclature',
                       #'pattern': re.compile(r'^\d+\.-\.-\.-|\d+\.\d+\.-\.-|\d+\.\d+\.\d+\.-|\d+\.\d+\.\d+\.(n)?\d+$'),
                       #'url': 'http://identifiers.org/ec-code/'}
               #})
MIRIAM_KEYS = list(MIRIAM)
MIRIAM_KEYS.sort()
MIRIAM_KEYS = tuple(MIRIAM_KEYS)
MIRIAM_KEYSlc = tuple([a.lower() for a in MIRIAM_KEYS])


from .CBConfig import __CBCONFIG__ as __CBCONFIG__
__DEBUG__ = __CBCONFIG__['DEBUG']
__version__ = __CBCONFIG__['VERSION']


cTime = time.time

class StructMatrix:
    """
    This class is specifically designed to store structural matrix information
    give it an array and row/col index permutations it can generate its own
    row/col labels given the label src.
    """

    array = None
    ridx = None
    cidx = None
    row = None
    col = None
    shape = None

    def __init__(self, array, ridx, cidx, row=None, col=None):
        """
        Instantiate with array and matching row/col index arrays, optional label arrays
        """
        self.array = array
        self.ridx = ridx
        self.cidx = cidx
        self.row = row
        self.col = col
        self.shape = array.shape

    def __call__(self):
        return self.array

    def getRowsByIdx(self, *args):
        """Return the rows referenced by index (1,3,5)"""
        return self.array.take(args, axis=0)

    def getColsByIdx(self, *args):
        """Return the columns referenced by index (1,3,5)"""
        return self.array.take(args, axis=1)

    def setRow(self, src):
        """
        Assuming that the row index array is a permutation (full/subset)
        of a source label array by supplying that source to setRow it
        maps the row labels to ridx and creates self.row (row label list)
        """
        self.row = [src[r] for r in self.ridx]

    def setCol(self, src):
        """
        Assuming that the col index array is a permutation (full/subset)
        of a source label array by supplying that src to setCol
        maps the row labels to cidx and creates self.col (col label list)
        """

        self.col = [src[c] for c in self.cidx]

    def getRowsByName(self, *args):
        """Return the rows referenced by label ('s','x','d')"""
        assert self.row != None, "\nI need row labels"
        try:
            return self.array.take([self.row.index(l) for l in args], axis=0)
        except Exception as ex:
            print(ex)
            print("\nValid row labels are: {}".format(self.row))
            return None

    def getColsByName(self, *args):
        """Return the columns referenced by label ('s','x','d')"""
        assert self.col != None, "\nI need column labels"
        try:
            return self.array.take([self.col.index(l) for l in args], axis=1)
        except Exception as ex:
            print(ex)
            print("Valid column labels are: {}".format(self.col))
            return None

    def getLabels(self, axis='all'):
        """Return the matrix labels ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row': return self.row
        elif axis == 'col': return self.col
        else: return self.row, self.col

    def getIndexes(self, axis='all'):
        """Return the matrix indexes ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row': return self.ridx
        elif axis == 'col': return self.cidx
        else: return self.ridx, self.cidx

    def getByIdx(self, row, col):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        return self.array[row, col]

    def getByName(self, row, col):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        return self.array[self.row.index(row), self.col.index(col)]

    def setByIdx(self, row, col, val):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        self.array[row, col] = val

    def setByName(self, row, col, val):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        self.array[self.row.index(row), self.col.index(col)] = val

class StructMatrixLP(StructMatrix):
    """Adds some stuff to StructMatrix that makes it LP friendly"""
    RHS = None
    operators = None
    __array_type__ = None

    def __init__(self, array, ridx, cidx, row=None, col=None, rhs=None, operators=None):
        """
        Object that holds an LP, stoichiometric constraints, operators, RHS, names etc

         - *array* stoichiometric matrix (linear constraints)
         - *ridx* a list of indexes typically range(shape[0])
         - *cidx* a list of indexes typically range(shape[1])
         - *row* a list of row names
         - *col* a list of col names
         - *rhs* [default=None] a list of rhs values defaults to 0
         - *operators* [default=None] a list of constraint senses defaults to 'E' =

        """
        StructMatrix.__init__(self, array, ridx, cidx, row=row, col=col)
        self.__array_type__ = type(array)
        if type(rhs) == type(None):
            self.RHS = numpy.zeros(array.shape[0])
        else:
            assert len(rhs) == array.shape[0], "\nRHS length mismatch"
            if type(rhs) == list:
                rhs = numpy.array(rhs)
            self.RHS = rhs
        if operators == None:
            self.operators = ['E']*len(self.RHS)
        else:
            assert len(operators) == len(self.RHS), "\nOperator length mismatch"
            self.operators = operators

    def getCopy(self, attr_str, deep=False):
        """
        Return a copy of the attribute with name attr_str. Uses the copy module `copy.copy` or `copy.deepcopy`

        - *attr_str* a string of the attribute name: 'row', 'col'
        - *deep* [default=False] try to do a deepcopy. Use with caution see copy module docstring for details

        """
        if attr_str == 'array' or attr_str == 'rhs':
            return getattr(self, attr_str).copy()
        elif deep:
            return copy.deepcopy(getattr(self, attr_str))
        else:
            return copy.copy(getattr(self, attr_str))

    def getLandRHS(self):
        RHS = self.RHS.copy()
        RHS.resize(len(RHS), 1)
        return numpy.hstack([self.array, RHS])

    def setRHSbyName(self, name, value):
        assert name in self.row, '\n%s is an invalid name' % name
        self.RHS[self.row.index(name)] = value

    def setRHSbyIdx(self, idx, val):
        assert idx in self.ridx, '\n%s is an invalid index' % idx
        self.RHS[idx] = val

    def setOperatorbyName(self, name, value):
        assert name in self.row, '\n%s is an invalid name' % name
        value = value.strip()
        assert value in ['G','L','E'], '\n%s is not a valid operator' % value
        self.operators[self.row.index(name)] = value

    def setOperatorbyIdx(self, idx, value):
        assert idx in self.ridx, '\n%s is an invalid index' % idx
        value = value.strip()
        assert value in ['G','L','E'], '\n%s is not a valid operator' % value
        self.operators[idx] = value

    def getRHSbyName(self, name):
        assert name in self.row, '\n%s is an invalid name' % name
        return self.RHS[self.row.index(name)]

    def getRHSbyIdx(self, idx):
        assert idx in self.ridx, '\n%s is an invalid index' % idx
        return self.RHS[idx]

    def getOperatorbyName(self, name):
        assert name in self.row, '\n%s is an invalid name' % name
        return self.operators[self.row.index(name)]

    def getOperatorbyIdx(self, idx):
        assert idx in self.ridx, '\n%s is an invalid index' % idx
        return self.operators[idx]

#TODO: this all needs to be redone to allow for nested qualifiers and model/biol qualifers

class MIRIAMannotation(object):
    """
    The MIRIAMannotation class MIRIAM annotations: Biological Qualifiers

    """

    MIRIAM = None
    MIDS = None
    MIDSlc = None
    QUALIFIERS = None
    # biomodels qualifiers
    isA = None
    isEncodedBy = None
    encodes = None
    hasPart = None
    hasProperty = None
    hasTaxon = None
    hasVersion = None
    isDescribedBy = None
    isHomologTo = None
    isPartOf = None
    isPropertyOf = None
    isVersionOf = None
    occursIn = None
    isDerivedFrom = None

    def __init__(self):
        self.MIRIAM = MIRIAM
        self.MIDS = MIRIAM_KEYS
        self.MIDSlc = MIRIAM_KEYSlc
        self.QUALIFIERS = ("isA","isEncodedBy","encodes","hasPart","hasProperty","hasTaxon","hasVersion","isDescribedBy",\
                           "isHomologTo","isPartOf","isPropertyOf","isVersionOf","occursIn",\
                           "isDerivedFrom")

    def addIDorgURI(self, qual, uri):
        """
        Add a URI directly into a qualifier collection:

         - *qual* a Biomodels biological qualifier e.g. "is" "isEncodedBy"
         - *uri* the complete identifiers.org uri e.g. http://identifiers.org/chebi/CHEBI:58088

        """
        if qual == 'is':
            qual = 'isA'
        if hasattr(self, qual):
            Q = self.__getattribute__(qual)
            if Q == None:
                Q = []
            Q.append(uri)
            self.__setattr__(qual, Q)
        else:
            print('INFO: Invalid qualifier: \"{}\" uri NOT set'.format(qual))
        #print(self.getAllMIRIAMUris())

    def checkEntityPattern(self, entity):
        """
        For an entity key compile the pattern to a regex, if necessary.

         - *entity* a MIRIAM resource entity

        """
        if type(self.MIRIAM[entity]['pattern']) == re._pattern_type:
            return True
        else:
            try:
                #self.MIRIAM[entity]['pattern'] = re.compile(self.MIRIAM[entity]['pattern'])
                return True
            except:
                print('Invalid pattern (entity={}): \"{}\"'.format(self.MIRIAM[entity]['name'], self.MIRIAM[entity]['pattern']))
                return False

    def checkId(self, entity, mid):
        """
        Check that a entity id e.g. CHEBI:17158

         - *mid* the entity id e.g. CHEBI:17158

        """
        if self.checkEntityPattern(entity):
            res = re.findall(self.MIRIAM[entity]['pattern'], mid)
            if len(res) > 0:
                return True
        print('INFO: invalid entity ({}) id: \"{}\"'.format(entity, mid))
        return False

    def checkEntity(self, entity):
        """
        Check an entity entry, this is a MIRIAM resource name: "chEBI". The test is case insensitive and will correct the case
        of wrongly capitalised entities automatically. If the entity is not recognised then a list of possible candidates
        based on the first letters of the input is displayed.

        - *entity* a MIRIAM resource entity e.g. "ChEBI"

        """
        if entity.lower() not in self.MIDSlc:
            print('\n\"{}\" is not a valid entity were you looking for one of these:\n'.format(entity))
            if len(entity) == 0:
                print('Need something to work with here')
            elif len(entity) == 1:
                temp = [a for a in self.MIDS if a[0].lower() == entity[0].lower()]
            else:
                temp = [a for a in self.MIDS if a[:2].lower() == entity[:2].lower()]
            for t_ in temp:
                print('\t{}'.format(t_))
            return None
        else:
            return self.MIDS[self.MIDSlc.index(entity.lower())]

    def addMIRIAMannotation(self, qual, entity, mid):
        """
        Add a qualified MIRIAM annotation or entity:

         - *qual* a Biomodels biological qualifier e.g. "is" "isEncodedBy"
         - *entity* a MIRIAM resource entity e.g. "ChEBI"
         - *mid* the entity id e.g. CHEBI:17158

        """
        if qual == 'is':
            qual = 'isA'
        if hasattr(self, qual):
            Q = self.__getattribute__(qual)
            if Q == None:
                Q = []
            E = self.checkEntity(entity)
            if E != None:
                if self.checkId(E, mid):
                    Q.append(self.MIRIAM[E]['url']+mid)
                    self.__setattr__(qual, Q)
            else:
                print('INFO: Invalid entity: \"{}\" MIRIAM entity NOT set'.format(entity))
        else:
            print('INFO: Invalid qualifier: \"{}\" MIRIAM entity NOT set'.format(qual))

    def deleteMIRIAMannotation(self, qual, entity, mid):
        """
        Deletes a qualified MIRIAM annotation or entity:

         - *qual* a Biomodels biological qualifier e.g. "is" "isEncodedBy"
         - *entity* a MIRIAM resource entity e.g. "ChEBI"
         - *mid* the entity id e.g. CHEBI:17158

        """
        if qual == 'is':
            qual = 'isA'
        if hasattr(self, qual):
            Q = self.__getattribute__(qual)
            if Q != None:
                E = self.checkEntity(entity)
                if E != None:
                    if self.checkId(E, mid):
                        annot = self.MIRIAM[E]['url']+mid
                        if annot in Q:
                            idx = Q.index(annot)
                            print('Deleted annotation {}'.format(Q.pop(idx)))
                        else:
                            print('INFO: Invalid entity: \"{}\" MIRIAM entry does not exist'.format(mid))
                else:
                    print('INFO: Invalid entity: \"{}\" MIRIAM entry does not exist'.format(mid))
            else:
                print('INFO: Invalid entity: \"{}\" MIRIAM entity does not exist'.format(entity))
        else:
            print('INFO: Invalid qualifier: \"{}\" MIRIAM entity does not exist'.format(qual))

    def getMIRIAMUrisForQualifier(self, qual):
        """
        Return all list of urls associated with qualifier:

         - *qual* the qualifier e.g. "is" or "isEncoded"

        """
        if qual == 'is':
            qual = 'isA'
        if hasattr(self, qual):
            Q = self.__getattribute__(qual)
            if Q == None:
                print('INFO: No \"{}\" qualifiers defined'.format(qual))
                return ()
            else:
                return tuple(Q)
        else:
            print('INFO: Invalid qualifier: \"{}\" MIRIAM entity NOT set'.format(qual))
            return ()

    def getAndViewUrisForQualifier(self, qual):
        """
        Retrieve all url's associated with qualifier and attempt to open them all in a new browser tab

         - *qual* the qualifier e.g. "is" or "isEncoded"

        """
        Q = self.getMIRIAMUrisForQualifier(qual)
        if Q != None:
            for q_ in Q:
                self.viewURL(q_)

    def getAllMIRIAMUris(self):
        """
        Return a dictionary of qualifiers that contain ID.org URL'S
        """
        out = {}
        for q_ in self.QUALIFIERS:
            Q = self.__getattribute__(q_)
            urls = []
            if Q != None:
                for u_ in self.getMIRIAMUrisForQualifier(q_):
                    urls.append(u_)
            if q_ == 'isA':
                q_ = 'is'
            out[q_] = tuple(urls)
        return out

    def viewURL(self, url):
        """
        This will try to open the URL in a new tab of the default webbrowser

         - *url* the url

        """
        webbrowser.open_new_tab(url)

#class MIRIAMModelAnnotation(MIRIAMannotation):
    #"""
    #Derived class with qualifiers for BQmodel
    #"""
    #isDerivedFrom = None


    #def __init__(self):
        #self.MIRIAM = MIRIAM
        #self.MIDS = MIRIAM_KEYS
        #self.MIDSlc = MIRIAM_KEYSlc
        #self.QUALIFIERS = ("isA", "isDerivedFrom", "isDescribedBy")



##  class SensitivityReaction(object):
    ##  """
    ##  Pseudo reaction class holding sensitivity data
    ##  """

    ##  def __init__(self):
        ##  pass

##  class SensitivitySpecies(object):
    ##  """
    ##  Pseudo species class holding sensitivity data
    ##  """

    ##  def __init__(self):
        ##  pass

##  class SensitivityData(object):
    ##  """
    ##  A class that holds sensitivity analysis data of an LP object
    ##  """

    ##  def __init__(self, sense_obj, sense_rhs, sense_bnd):
        ##  pass

    ##  def getListOfReactions(self):
        ##  """
        ##  Returns a list of sensitivity reaction objects
        ##  """
        ##  return []

    ##  def getListOfConstraints(self):
        ##  """
        ##  Returns a list of sensitivity constraint objects
        ##  """
        ##  return []



"""
CBMPy: CBConfig module
======================
PySCeS Constraint Based Modelling (http://cbmpy.sourceforge.net)
Copyright (C) 2009-2018 Brett G. Olivier, VU University Amsterdam, Amsterdam, The Netherlands

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>

Author: Brett G. Olivier
Contact email: bgoli@users.sourceforge.net
Last edit: $Author: bgoli $ ($Id: CBConfig.py 660 2018-09-24 14:57:04Z bgoli $)

"""
## gets rid of "invalid variable name" info
# pylint: disable=C0103
## gets rid of "line to long" info
# pylint: disable=C0301
## use with caution: gets rid of module xxx has no member errors (run once enabled)
# pylint: disable=E1101

# preparing for Python 3 port
from __future__ import division, print_function
from __future__ import absolute_import
#from __future__ import unicode_literals

# release

try:
    STATUS = '$Rev: 660 $'.replace('Rev: ', '').replace('$', '').strip()
except Exception:
    STATUS = ''

__CBCONFIG__ = {'VERSION_MAJOR' : 0,
                'VERSION_MINOR' : 7,
                'VERSION_MICRO' : 21,
                'VERSION_STATUS' : STATUS,
                'VERSION' : None,
                'DEBUG' : False,
                'SOLVER_PREF' : 'CPLEX',
                #'SOLVER_PREF' : 'GLPK',
                'SOLVER_ACTIVE' : None,
                'REVERSIBLE_SYMBOL' : '<==>',
                'IRREVERSIBLE_SYMBOL' : '-->',
                'HAVE_SBML2' : False,
                'HAVE_SBML3' : False,
                'CBMPY_DIR' : None,
                'CBMPY_CGLPK_VER' : 'w452',
                'SYMPY_DENOM_LIMIT' : 10**32
               }

def current_version():
    """
    Return the current CBMPy version as a string

    """
    #return '%s.%s.%s [r%s]' % (__CBCONFIG__['VERSION_MAJOR'], __CBCONFIG__['VERSION_MINOR'], __CBCONFIG__['VERSION_MICRO'], __CBCONFIG__['VERSION_STATUS'])
    if STATUS == '':
        return '{}.{}.{}'.format(__CBCONFIG__['VERSION_MAJOR'], __CBCONFIG__['VERSION_MINOR'], __CBCONFIG__['VERSION_MICRO'])
    else:
        return '{}.{}.{}.{}'.format(__CBCONFIG__['VERSION_MAJOR'], __CBCONFIG__['VERSION_MINOR'], __CBCONFIG__['VERSION_MICRO'], STATUS)

def current_version_tuple():
    """
    Return the current CBMPy version as a tuple (x, y, z)

    """
    return (__CBCONFIG__['VERSION_MAJOR'], __CBCONFIG__['VERSION_MINOR'], __CBCONFIG__['VERSION_MICRO'])

__CBCONFIG__['VERSION'] = current_version()


